/*
This file is licensed under the terms of the Globus Toolkit Public
License, found at http://www.globus.org/toolkit/download/license.html.
*/
package org.globus.ftp;

import java.io.IOException;
import java.io.File;
import java.io.RandomAccessFile;
import java.util.Vector;
import java.net.UnknownHostException;

import org.globus.ftp.exception.ClientException;
import org.globus.ftp.exception.ServerException;
import org.globus.ftp.exception.FTPReplyParseException;
import org.globus.ftp.exception.UnexpectedReplyCodeException;
import org.globus.ftp.vanilla.Command;
import org.globus.ftp.vanilla.Reply;
import org.globus.ftp.extended.GridFTPServerFacade;
import org.globus.ftp.extended.GridFTPControlChannel;
import org.globus.gsi.gssapi.auth.Authorization;

import org.ietf.jgss.GSSCredential;

import org.apache.log4j.Logger;

/**
 * This is the main user interface for GridFTP operations.
 * Use this class for client - server or third party transfers
 * with mode E, parallelism, markers, striping or GSI authentication.
 * Consult the manual for general usage.
 * <br><b>Note:</b> If using with GridFTP servers operations like
 * {@link #setMode(int) setMode()}, {@link #setType(int) setType()},
 * {@link #setDataChannelProtection(int) setDataChannelProtection()},
 * and {@link #setDataChannelAuthentication(DataChannelAuthentication)
 * setDataChannelAuthentication()} that affect data channel settings 
 * <b>must</b> be called before passive or active data channel mode is set. 
 **/
public class GridFTPClient extends FTPClient {

    private static Logger logger = 
	Logger.getLogger(GridFTPClient.class.getName());

    //utility alias to session and localServer
    protected GridFTPSession gSession;
    protected GridFTPServerFacade gLocalServer;

    /**
     * Constructs client and connects it to the remote server.
     * 
     * @param host remote server host
     * @param port remote server port
     */
    public GridFTPClient(String host, int port) 
	throws IOException, ServerException {

	gLocalServer = new GridFTPServerFacade();
	localServer = gLocalServer;
	gLocalServer.authorize();

	gSession = new GridFTPSession();
	session = gSession;

	controlChannel = new GridFTPControlChannel(host, port);
	controlChannel.open();
    }

    /**
     * Performs authentication with specified user credentials.
     *
     * @param credential user credentials to use.
     * @throws IOException on i/o error
     * @throws ServerException on server refusal or faulty server behavior
     */
    public void authenticate( GSSCredential credential )
	throws IOException,
	       ServerException{
	((GridFTPControlChannel)controlChannel).authenticate(credential);
	gLocalServer.setCredential(credential);
	gSession.authorized = true;
    }

    /**
     * Performs remote directory listing, like FTPClient.list(). 
     * <b>Note:</b> This method cannot be used
     * in conjunction with parallelism or striping; set parallelism to
     * 1 before calling it. Otherwise, use the parametrized list().
     * Unlike in vanilla FTP, here IMAGE mode is allowed. 
     * For more documentation, look at FTPClient.
     */
    public Vector list() 
	throws  ServerException,
		ClientException,
		IOException {
	if (gSession.parallel > 1) {
	    throw new ClientException(
				      ClientException.BAD_MODE,
				      "list() cannot be called with parallelism");
	}
	return super.list();
    }

    /**
     * Use this method to list remote directory while in parallel transfer
     * mode. The DataSink you support must be thread safe.
     * Look at FTPClient for more documentation.
     **/
    public void list(String filter,
		     String modifier,
		     DataSink sink) 
	throws  ServerException,
		ClientException,
		IOException {
	logger.debug("GridFTPClient parametrized list");
	// all parameters set correctly (or still unset)?
	gSession.matches(localServer.getSession());

	// if transfer modes not defined,
	// set remote server (dest) as passive
	if (gSession.serverMode == gSession.SERVER_DEFAULT) {
	    HostPort hp = setLocalPassive();
	    this.setActive(hp);
	}

	actualList(filter, modifier, sink);

    }

    protected void listCheck() throws ClientException {
	// do nothing
    }

    /**
     * Sets transfer mode.
     * @param mode should be {@link Session#MODE_STREAM MODE_STREAM}, 
     *                       {@link Session#MODE_BLOCK MODE_BLOCK}, or 
     *                       {@link GridFTPSession#MODE_EBLOCK MODE_EBLOCK}
     **/
    public void setMode(int mode) 
	throws IOException, 
	       ServerException { 

	String modeStr = null;
	switch (mode) {
	case Session.MODE_STREAM: modeStr = "S"; 
	    break;
	case Session.MODE_BLOCK: modeStr = "B";
	    break;
	case GridFTPSession.MODE_EBLOCK: modeStr = "E";
	    break;
	default: throw new IllegalArgumentException("Bad mode"); 
	}
	
	actualSetMode(mode, modeStr);
    }

    /**
     * Sets remote server TCP buffer size, in the following way:
     * First see if server supports "SBUF" and if so, use it.
     * If not, try the following commands until success:
     * "SITE RETRBUFSIZE", "SITE RBUFSZ", "SITE RBUFSIZ", 
     * "SITE STORBUFIZE",  "SITE SBUFSZ",  "SITE SBUFSIZ",
     * "SITE BUFSIZE".
     * Returns normally if the server confirms successfull setting of the 
     * remote buffer size, both for sending and for receiving data.
     * Otherwise, throws ServerException.
     **/
    public void setTCPBufferSize(int size) 
	throws IOException, 
	       ServerException {
	try {
	    if (size <=0 ) {
		throw new IllegalArgumentException("size <= 0");
	    }
	    
	    boolean succeeded = false;
	    String sizeString = Integer.toString(size);
	    FeatureList feat = getFeatureList();

	    if (feat.contains(FeatureList.SBUF)) {
		succeeded = tryExecutingCommand( new Command("SBUF", sizeString));
	    }
	    
	    if (!succeeded) {
		succeeded = tryExecutingCommand( new Command("SITE BUFSIZE", sizeString));
	    }
	    
	    if (!succeeded) {
		succeeded = tryExecutingTwoCommands(new Command("SITE RETRBUFSIZE", sizeString),
						    new Command("SITE STORBUFIZE", sizeString));
	    }
	    
	    if (!succeeded) {
		succeeded = tryExecutingTwoCommands(new Command("SITE RBUFSZ", sizeString),
						    new Command("SITE SBUFSZ", sizeString));
	    }
	    
	    if (!succeeded) {
		succeeded = tryExecutingTwoCommands(new Command("SITE RBUFSIZ", sizeString),
						    new Command("SITE SBUFSIZ", sizeString));
	    }
	    
	    if (succeeded) {
		this.gSession.TCPBufferSize = size;
	    } else {
		throw new ServerException(ServerException.SERVER_REFUSED,
					  "Server refused setting TCP buffer size with any of the known commands.");
	    }

	} catch (FTPReplyParseException rpe) {
	    throw ServerException.embedFTPReplyParseException(rpe);
	}
    }

    private boolean tryExecutingTwoCommands(Command cmd1, Command cmd2) 
	throws IOException, 
	       FTPReplyParseException, 
	       ServerException {
	boolean result = tryExecutingCommand(cmd1);
	if (result) {
	    result = tryExecutingCommand(cmd2);
	}
	return result;
    }

    /*
     * This is like controlChannel.executeCommand, only that negative reply it
     * returns "false" rather than throwing exception
     */
    private boolean tryExecutingCommand(Command cmd) 
	throws IOException, 
	       FTPReplyParseException, 
	       ServerException {
	Reply reply = controlChannel.exchange(cmd);
	return Reply.isPositiveCompletion(reply);
    }
		
    /**
     * Sets local TCP buffer size (for both receiving and sending).
     **/
    public void setLocalTCPBufferSize(int size) 
	throws ClientException {
	if (size <=0 ) {
	    throw new IllegalArgumentException("size <= 0");
	}
	gLocalServer.setTCPBufferSize(size);
	
    }
    
    /**
     * Sets remote server to striped passive server mode (SPAS).
     **/
    public HostPortList setStripedPassive() 
	throws IOException, 
	       ServerException {
	    Reply reply = null;

	    try {
		reply = controlChannel.execute(Command.SPAS);
	    } catch (UnexpectedReplyCodeException urce) {
		throw ServerException.embedUnexpectedReplyCodeException(urce);
	    } catch(FTPReplyParseException rpe) {
		throw ServerException.embedFTPReplyParseException(rpe);
	    }

	    this.gSession.serverMode = gSession.SERVER_EPAS;
	    gSession.serverAddressList = new HostPortList(reply.getMessage());
	    return gSession.serverAddressList;
    }
    
    /**
     * Sets remote server to striped active server mode (SPOR).
     **/
    public void setStripedActive(HostPortList hpl)
	throws IOException, 
	       ServerException {
	Command cmd = new Command("SPOR", hpl.toFtpCmdArgument());

	try {
	    controlChannel.execute(cmd);
	} catch (UnexpectedReplyCodeException urce) {
	    throw ServerException.embedUnexpectedReplyCodeException(urce);
	} catch(FTPReplyParseException rpe) {
	    throw ServerException.embedFTPReplyParseException(rpe);
	}
	
	this.gSession.serverMode = gSession.SERVER_EACT;
    }
    
    /** 
     * Starts local server in striped passive mode. Since the local server
     * is not distributed, it will only listen on one socket.
     *
     * @param port required server port; can be set to FTPServerFacade.ANY_PORT
     * @param queue max size of queue of awaiting new data channel connection
     *              requests
     * @return the HostPortList of 1 element representing the socket where the
     *         local server is listening
     **/
    public HostPortList setLocalStripedPassive(int port, int queue) 
	throws IOException {
	return gLocalServer.setStripedPassive(port, queue);
    }

    /**
     * Behaves like setLocalStripedPassive(FTPServerFacade.ANY_PORT, 
     * FTPServerFacade.DEFAULT_QUEUE)
     **/
    public HostPortList setLocalStripedPassive() 
	throws IOException {
	return gLocalServer.setStripedPassive();
    }

    /** 
     * Starts local server in striped active mode.
     * setStripedPassive() must be called before that.
     * This method takes no parameters. HostPortList of the remote
     * server, known from the last call of setStripedPassive(), is stored
     * internally and the local server will connect to this address.
     **/
    public void setLocalStripedActive() 
	throws ClientException, 
	       IOException {
	if (gSession.serverAddressList == null) {
	    throw new ClientException(ClientException.CALL_PASSIVE_FIRST);
	}
	try {
	    gLocalServer.setStripedActive(gSession.serverAddressList);
	} catch (UnknownHostException e) {
	    throw new ClientException(ClientException.UNKNOWN_HOST);
	}
    }

    /**
     * Performs extended retrieve (partial retrieve mode starting
     * at offset 0).
     *
     * @param remoteFileName file to retrieve
     * @param size number of bytes of remote file to transmit
     * @param sink data sink to store the file
     * @param mListener marker listener
     **/
    public void extendedGet(String remoteFileName,
			    long size,
			    DataSink sink,
			    MarkerListener mListener)
	throws IOException,
	       ClientException,
	       ServerException {
	extendedGet(remoteFileName,
		    0,
		    size,
		    sink,
		    mListener);
    }

    /**
     * Performs extended retrieve (partial retrieve mode).
     *
     * @param remoteFileName file to retrieve
     * @param offset the staring offset in the remote file
     * @param size number of bytes of remote file to transmit
     * @param sink data sink to store the file
     * @param mListener marker listener
     **/
    public void extendedGet(String remoteFileName,
			    long offset,
			    long size,
			    DataSink sink,
			    MarkerListener mListener)
	throws IOException,
	       ClientException,
	       ServerException {
	
	// servers support GridFTP?
	checkGridFTPSupport();
	// all parameters set correctly (or still unset)?
	gSession.matches(localServer.getSession());
	
	// if transfer modes not defined,
	// set remote server (source) as active
	if (gSession.serverMode == gSession.SERVER_DEFAULT) {
	    HostPort hp = setLocalPassive();
	    this.setActive(hp);
	}

	logger.debug("localServer.store()");
	gLocalServer.store(sink);

	controlChannel.write(new Command("ERET", 
					 "P " + offset + " " + size
					 + " " + remoteFileName));
	
	transferRun( localServer.getControlChannel(),
		     mListener);
    }

    /**
     * Performs extended store (adujsted store mode with offset 0).
     * 
     * @param remoteFileName file name to store
     * @param source source for the data to transfer
     * @param mListener marker listener
     **/
    public void extendedPut(String remoteFileName,
			    DataSource source,
			    MarkerListener mListener)
	throws IOException,
	       ServerException,
	       ClientException{
	extendedPut(remoteFileName,
		    0,
		    source,
		    mListener);
    }

    /**
     * Performs extended store (adujsted store mode).
     * 
     * @param remoteFileName file name to store
     * @param offset the offset added to the file pointer before storing
     *               the blocks of the file.
     * @param source source for the data to transfer
     * @param mListener marker listener
     **/
    public void extendedPut(String remoteFileName,
			    long offset,
			    DataSource source,
			    MarkerListener mListener)
	throws IOException,
	       ServerException,
	       ClientException{

	// servers support GridFTP?
	checkGridFTPSupport();
	// all parameters set correctly (or still unset)?
	gSession.matches(localServer.getSession());
	
	// if transfer modes have not been defined, 
	// set this (source) as active
	if (gSession.serverMode == gSession.SERVER_DEFAULT) {
	    HostPort hp = this.setPassive();
	    setLocalActive();
	}

	logger.debug("localServer.retrieve()");
	localServer.retrieve(source);

	controlChannel.write(new Command("ESTO",
					 "A " + offset + " " + 
					 remoteFileName));
	
	transferRun(localServer.getControlChannel(),
		    mListener);			     
    }

    /*
      3rd party transfer code
     */


    /**
     * Performs a third-party transfer between two servers using extended 
     * block mode.
     * If server modes are unset, source will be set to active
     * and destination to passive.
     *
     * @param remoteSrcFile source filename
     * @param dstServer     destination server
     * @param remoteDstFile destination filename
     * @param append        enables append mode, if true
     *                      file will be appened.
     * @param listener      transer progress listener.
     *                      Can be set to null.
     */
    public void extendedTransfer(String remoteSrcFile, 
				 GridFTPClient destination, 
				 String remoteDstFile,
				 MarkerListener mListener)
	throws IOException, 
	       ServerException,
	       ClientException {

	// FIXME: ESTO & ERET do not require MODE E this needs to be fixed

	// servers support GridFTP?
	checkGridFTPSupport();
	destination.checkGridFTPSupport();
	// all parameters set correctly (or still unset)?
	gSession.matches(destination.gSession);
	
	//mode E
	if (gSession.transferMode != GridFTPSession.MODE_EBLOCK) {
	    throw new ClientException(ClientException.BAD_MODE,
	        "Extended transfer mode is necessary");
	}

	// if transfer modes have not been defined, 
	// set this (source) as active
	if (gSession.serverMode == gSession.SERVER_DEFAULT) {
	    
	    HostPort hp = destination.setPassive();
	    this.setActive(hp);
	    
	} 

	long sourceFileSize = getSize(remoteSrcFile);

	destination.controlChannel.write( new Command(
						      "ESTO",
						      "A 0 " + remoteDstFile
 ));
	controlChannel.write(new Command("ERET", 
					"P 0 " + sourceFileSize
					+ " " + remoteSrcFile));

	transferRun(destination.controlChannel,
		    mListener);
    }

    /**
     * assure that the server supports extended transfer features;
     * throw exception if not
     **/
    protected void checkGridFTPSupport() 
	throws IOException,
	       ServerException {	    
	FeatureList fl = getFeatureList();
	if (
	    !(fl.contains(FeatureList.PARALLEL)
	      && fl.contains(FeatureList.ESTO)
	      && fl.contains(FeatureList.ERET)
	      && fl.contains(FeatureList.SIZE))
	    ) {
	    throw new ServerException(ServerException.UNSUPPORTED_FEATURE);
	}
	logger.debug("okay, server supports extended, parallel and striped");
    }

    /*
     end 3rd party transfer code
    */

    /**
     * Sets data channel authentication mode (DCAU)
     * 
     * @param type; for 2-party transfer must be
     * DataChannelAuthentication.SELF or DataChannelAuthentication.NONE
     **/
    public void setDataChannelAuthentication(DataChannelAuthentication type) 
	throws IOException, 
	       ServerException {

	Command cmd = new Command("DCAU", type.toFtpCmdArgument());
	try{

	    controlChannel.execute(cmd);

	} catch (UnexpectedReplyCodeException urce) {
	    throw ServerException.embedUnexpectedReplyCodeException(urce);
	} catch(FTPReplyParseException rpe) {
	    throw ServerException.embedFTPReplyParseException(rpe);
	}

	this.gSession.dataChannelAuthentication = type;

	gLocalServer.setDataChannelAuthentication(type);
    }
    
    /**
     * Sets compatibility mode with old GSIFTP server.
     * Locally sets data channel authentication to NONE 
     * but does not send the command
     * to the remote server (the server wouldn't understand it)
     **/
    public void setLocalNoDataChannelAuthentication() {
	gLocalServer.setDataChannelAuthentication(DataChannelAuthentication.NONE);
    }

    /**
     * Returns data channel authentication mode (DCAU).
     * 
     * @return data channel authentication mode
     **/
    public DataChannelAuthentication getDataChannelAuthentication() {
	return gSession.dataChannelAuthentication;
    }

    /**
     * Sets data channel protection level (PROT).
     * @param protection should be 
     *             {@link GridFTPSession#PROTECTION_CLEAR CLEAR},
     *             {@link GridFTPSession#PROTECTION_SAFE SAFE}, or
     *             {@link GridFTPSession#PROTECTION_PRIVATE PRIVATE}, or
     *             {@link GridFTPSession#PROTECTION_CONFIDENTIAL CONFIDENTIAL}.
     **/
    public void setDataChannelProtection(int protection)
	throws IOException,
	       ServerException {

	String protectionStr = null;
	switch(protection) {
	case GridFTPSession.PROTECTION_CLEAR:
	    protectionStr = "C"; break;
	case GridFTPSession.PROTECTION_SAFE:
	    protectionStr = "S"; break;
	case GridFTPSession.PROTECTION_CONFIDENTIAL:
	    protectionStr = "E"; break;
	case GridFTPSession.PROTECTION_PRIVATE:
	    protectionStr = "P"; break;
	default: throw new IllegalArgumentException("Bad protection: " +
						    protection);
	}
	
	Command cmd = new Command("PROT", protectionStr);
	try {

	    controlChannel.execute(cmd);

	} catch (UnexpectedReplyCodeException urce) {
	    throw ServerException.embedUnexpectedReplyCodeException(urce);
	} catch(FTPReplyParseException rpe) {
	    throw ServerException.embedFTPReplyParseException(rpe);
	}

	this.gSession.dataChannelProtection = protection;

	gLocalServer.setDataChannelProtection(protection);	
    }
    
    /**
     * Returns data channel protection level.
     * 
     * @return data channel protection level: 
     *             {@link GridFTPSession#PROTECTION_CLEAR CLEAR},
     *             {@link GridFTPSession#PROTECTION_SAFE SAFE}, or
     *             {@link GridFTPSession#PROTECTION_PRIVATE PRIVATE}, or
     *             {@link GridFTPSession#PROTECTION_CONFIDENTIAL CONFIDENTIAL}.
     **/
    public int getDataChannelProtection() {
	return gSession.dataChannelProtection;
    }

    /**
     * Sets authorization method for the control channel.
     *
     * @param authorization authorization method.
     */
    public void setAuthorization(Authorization authorization) {
	((GridFTPControlChannel)this.controlChannel).setAuthorization(authorization);
    }
    
    /**
     * Returns authorization method for the control channel.
     * 
     * @return authorization method performed on the control channel.
     */
    public Authorization getAuthorization() {
	return ((GridFTPControlChannel)this.controlChannel).getAuthorization();
    }



    // basic compatibility API

    public void get(String remoteFileName,
		    File localFile) 
	throws IOException,
	       ClientException,
	       ServerException {
	if (gSession.transferMode == GridFTPSession.MODE_EBLOCK) {
	    DataSink sink = 
		new FileRandomIO(new RandomAccessFile(localFile, "rw"));
	    get(remoteFileName, sink, null);
	} else {
	    super.get(remoteFileName, localFile);
	}
    }

    public void put(File localFile,
		    String remoteFileName,
		    boolean append) 
	throws IOException,
	       ServerException,
	       ClientException{
	if (gSession.transferMode == GridFTPSession.MODE_EBLOCK) {
	    DataSource source = 
		new FileRandomIO(new RandomAccessFile(localFile, "r"));
	    put(remoteFileName, source, null, append);
	} else {
	    super.put(localFile, remoteFileName, append);
	}
    }
}
