/*
This file is licensed under the terms of the Globus Toolkit Public
License, found at http://www.globus.org/toolkit/download/license.html.
*/
package org.globus.ftp;

import java.util.Vector;
import java.util.StringTokenizer;

/**
 * Utility class for parsing
 * and converting host-port information from SPAS 
 * and SPOR FTP commands. Represents a list of host-port pairs.
 */
public class HostPortList {
    /*
      This class is internally represented as String or as vector.
      In case of third party transfer, there is no need to convert the String
      to vector, since we receive a string from server A and send a similar
      string to server B. However in client-server transfer, we do need internal
      vector representation.
      Internally, if the constructor gets string as parameter, the default 
      representation remains string. Whenever you call a modifier method like
      add(), the string becomes out of date and then the vector is used.
      The string can become up to date again if updateString() called.
      The vector is not usable until updateVector() is first called,
      and then it always remains up to date.
     */
    
    static String nl = System.getProperty("line.separator");
  
    protected final static boolean 
	UP_TO_DATE = true,
	OUT_OF_DATE = false;

    private boolean stringStatus = OUT_OF_DATE;
    //internal string, in form of parameters to SPOR command
    private String sporCommandParam;
    private Vector vector = null;
    /**
     * Parses host-port from the reply to SPAS command.
     *
     * @param passiveReply reply message for the SPAS command
     */
    public HostPortList(String spasReplyMsg) {
	    sporCommandParam = "";
	    //previous line's CRLF is at..
	    int previousEnds = 0;

	    while (true) {
		previousEnds = spasReplyMsg.indexOf(nl,
						    previousEnds + 1);
		int currentEnds = spasReplyMsg.indexOf(nl,
						    previousEnds + 1);
		if (currentEnds == -1) {
		    currentEnds = spasReplyMsg.length();
		}
		// newline length is OS specific 
		String currentLine = spasReplyMsg.substring(previousEnds + nl.length(), 
							    currentEnds);

		//last line
		if (!currentLine.startsWith(" ")) {
		    if (currentLine.startsWith("229")) {
			break;
		    } else {
			throw new IllegalArgumentException(
			   "Argument to HostPortList constructor is not " + 
			   "a 229 reply message: last line " +
			   "->" + currentLine  + "<-");
		    }    
	    }
		//make sure there's just 1 space
		currentLine = currentLine.trim();
		sporCommandParam += " " + currentLine;
	    }
	    //remove leading space ' '
	    sporCommandParam = sporCommandParam.substring(1, sporCommandParam.length());

	    stringStatus = UP_TO_DATE;
    }


    /**
       Creates an empty list
     **/
    public HostPortList() {
	stringStatus = OUT_OF_DATE;
    }


    protected void updateString() {
	// start from scratch
	sporCommandParam = "";
	for (int i = 0; i < vector.size(); i ++) {
	    HostPort hp = (HostPort)vector.get(i);
	    if (i != 0) {
		sporCommandParam += " ";
	    }
	    sporCommandParam +=  hp.toFtpCmdArgument();
	}
    }

    // update the internal list representation
    // from flat (string) to to vector of objects
    protected void updateVector() {
	if (vector == null) parse();
    }

    private void parse() {
	vector = new Vector();
	if (sporCommandParam != null) {
	    StringTokenizer tokens = new StringTokenizer(sporCommandParam, 
							 nl);
	    while(tokens.hasMoreTokens()) {
		vector.add(new HostPort(tokens.nextToken()));
	    }
	}
    }

    /**
       Adds an element to the list
     **/
    public void add(HostPort hp) {
	updateVector();
	vector.add(hp);
	stringStatus = OUT_OF_DATE;
    }
    
    /**
       @return number of elements in the list
     **/
    public int size() {
	updateVector();
	return vector.size();
    }
    /**
       @return element of the specified index
     **/
    public HostPort get(int index) {
	updateVector();
	return (HostPort)vector.elementAt(index);
    }

    /**
     * Returns the host-port infromation in the
     * format used by SPOR command. 
     *
     * @return host-port information in SPOR command parameter
     *         representation.
     */
    public String toFtpCmdArgument() {
	if (stringStatus == OUT_OF_DATE) {
	    updateString();
	}
	    return sporCommandParam;
    }
    
}
