/*
This file is licensed under the terms of the Globus Toolkit Public
License, found at http://www.globus.org/toolkit/download/license.html.
*/
package org.globus.ftp.extended;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import org.globus.ftp.exception.ServerException;
import org.globus.ftp.exception.UnexpectedReplyCodeException;
import org.globus.ftp.exception.FTPReplyParseException;
import org.globus.ftp.vanilla.Reply;
import org.globus.ftp.vanilla.FTPControlChannel;
import org.globus.ftp.vanilla.Command;
import org.globus.common.ChainedIOException;
import org.globus.gsi.gssapi.auth.Authorization;
import org.globus.gsi.gssapi.auth.GSSAuthorization;
import org.globus.gsi.gssapi.auth.HostAuthorization;
import org.globus.gsi.gssapi.auth.AuthorizationException;
import org.globus.gsi.gssapi.GSSConstants;

import org.gridforum.jgss.ExtendedGSSManager;

import org.ietf.jgss.GSSManager;
import org.ietf.jgss.GSSContext;
import org.ietf.jgss.GSSException;
import org.ietf.jgss.GSSCredential;
import org.ietf.jgss.GSSName;

import org.apache.log4j.Logger;

/** GridFTP control channel, unlike the vanilla control channel,
    uses GSI autentication.
 **/
public class GridFTPControlChannel extends FTPControlChannel {

    private static Logger logger = 
	Logger.getLogger(GridFTPControlChannel.class.getName());

    protected static final int TIMEOUT = 120000;

    //maybe this is useless
    protected GSSCredential credentials = null;

    protected Authorization authorization = 
	HostAuthorization.getInstance();

    public GridFTPControlChannel(String host, int port) {
	super(host,port);
    }

    public GridFTPControlChannel(InputStream in, OutputStream out) {
	super(in, out);
    }

    /**
     * Sets authorization method for the control channel.
     *
     * @param authorization authorization method.
     */
    public void setAuthorization(Authorization authorization) {
	this.authorization = authorization;
    }

    /**
     * Returns authorization method for the control channel.
     * 
     * @return authorization method performed on the control channel.
     */
    public Authorization getAuthorization() {
	return this.authorization;
    }

   /**
     * Performs authentication with specified user credentials.
     *
     * @param credential user credentials to use.
     * @throws IOException on i/o error
     * @throws ServerException on server refusal or faulty server behavior
     */
    public void authenticate( GSSCredential credential )
	    throws IOException,
		   ServerException{
	
	setCredentials( credential );
	 
	write(new Command("AUTH", "GSSAPI"));

	Reply reply0 = null;
	try {
	    reply0 = read();
	} catch (FTPReplyParseException rpe) {
	    throw ServerException.embedFTPReplyParseException(
				      rpe,
				      "Received faulty reply to AUTH GSSAPI");
	}

	if (! Reply.isPositiveIntermediate(reply0)) {
	   close();
	   throw ServerException.embedUnexpectedReplyCodeException(
				  new UnexpectedReplyCodeException(reply0),   
				  "Server refused GSSAPI authentication.");
	}

	GSSManager manager = ExtendedGSSManager.getInstance();

	GSSContext context = null;
	GridFTPOutputStream gssout = null;
	GridFTPInputStream gssin = null;

	try {
            String host = this.socket.getInetAddress().getHostAddress();

            GSSName expectedName = null;

            if (this.authorization instanceof GSSAuthorization) {
                GSSAuthorization auth = (GSSAuthorization)this.authorization;
                expectedName = auth.getExpectedName(credential, host);
            }

	    context = manager.createContext(expectedName,
					    GSSConstants.MECH_OID,
					    credential,
					    GSSContext.DEFAULT_LIFETIME);
	    context.requestCredDeleg(true);
	        
	    gssout = new GridFTPOutputStream(ftpOut, context);
	    gssin = new GridFTPInputStream(rawFtpIn, context);

	    byte [] inToken = new byte[0];
	    byte [] outToken = null;
	    
	    while( !context.isEstablished() ) {
		
		outToken = context.initSecContext(inToken, 0, inToken.length);
		
		if (outToken != null) {
		    gssout.writeHandshakeToken(outToken);
		}

		if (!context.isEstablished()) {
		    inToken = gssin.readHandshakeToken();
		}
	    }

	} catch (GSSException e) {
	    throw new ChainedIOException("Authentication failed", e);
	}

	if (this.authorization != null) {
	    try {
		this.authorization.authorize(context, host);
	    } catch (AuthorizationException e) {
		throw new ChainedIOException("Authorization failed", e);
	    }
	}

	// this should be authentication success msg (plain)
	// 234 (ok, no further data required)
	Reply reply1 = null;
	try {
	    reply1    = read();
	} catch (FTPReplyParseException rpe) {
	    throw ServerException.embedFTPReplyParseException(
				      rpe,
				      "Received faulty reply to authentication");

	}
	
	if ( ! Reply.isPositiveCompletion(reply1)) {
	    close();
	    throw ServerException.embedUnexpectedReplyCodeException(
				    new UnexpectedReplyCodeException(reply1),
				    "GSSAPI authentication failed.");
	}
	
	// enter secure mode - send MIC commands
	setInputStream(gssin);
	setOutputStream(gssout);
	//from now on, the commands and replies
	//are protected and pass through gsi wrapped socket

	write(new Command("USER", ":globus-mapping:"));

	Reply reply2 = null;
	try {
	    reply2 = read();
	} catch (FTPReplyParseException rpe) {
	    throw ServerException.embedFTPReplyParseException(
				      rpe,
				      "Received faulty reply to USER command");
	}


	if (Reply.isPositiveCompletion(reply2) ||
	    Reply.isPositiveIntermediate(reply2)) {
	   // wu-gsiftp sends intermediate code while
	   // gssftp send completion reply code
	} else {
	   close();
	   throw ServerException.embedUnexpectedReplyCodeException(
				        new UnexpectedReplyCodeException(reply2),
					"User authorization failed.");
	}


	write(new Command("PASS", "dummy"));

	Reply reply3 = null;
	try {
	    reply3=read();
	} catch (FTPReplyParseException rpe) {
	    throw ServerException.embedFTPReplyParseException(
				     rpe,
				     "Received faulty reply to PASS command");
	}

	if (!Reply.isPositiveCompletion(reply3)) {
	    close();
	    throw ServerException.embedUnexpectedReplyCodeException(
				    new UnexpectedReplyCodeException(reply3),
			      	    "Bad password.");
	}
    }

    protected void setCredentials( GSSCredential credentials ) {
	this.credentials = credentials;
    }

    protected GSSCredential getCredentials() {
	return credentials;
    }

}








