/*
This file is licensed under the terms of the Globus Toolkit Public
License, found at http://www.globus.org/toolkit/download/license.html.
*/
package org.globus.gsi;

import java.security.Security;
import java.security.Provider;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.GeneralSecurityException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.cert.X509CRL;
import java.security.Principal;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.InputStream;
import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.util.StringTokenizer;

import org.globus.util.Base64;
import org.globus.common.CoGProperties;
import org.globus.util.log4j.NullAppender;
import org.globus.gsi.bc.X509NameHelper;

import org.apache.log4j.Logger;

import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.asn1.x509.X509Name;

/** 
 * Contains various security-related utility methods.
 */
public class CertUtil {
  
    private static Logger logger =
	Logger.getLogger(CertUtil.class.getName());
    
    static {
	// this is a hack to prevent the log4j warning messages
	Logger.getRootLogger().addAppender(NullAppender.getInstance());
	Security.addProvider(new BouncyCastleProvider());
	setProvider("BC");
	installSecureRandomProvider();
    }

    private static String provider;
    private static CertificateFactory certFact;

    /**
     * A no-op function that can be used to force the class
     * to load and initialize.
     */
    public static void init() {
    }

    /**
     * Sets a provider name to use for loading certificates
     * and for generating key pairs.
     *
     * @param providerName provider name to use.
     */
    public static void setProvider(String providerName) {
	provider = providerName;
	if (logger.isDebugEnabled()) {
	    logger.debug("Provider set to : " + providerName);
	}
    }

    /**
     * Returns appropriate <code>CertificateFactory</code>.
     * If <I>provider</I> was set a provider-specific 
     * <code>CertificateFactory</code> will be used. Otherwise, 
     * a default <code>CertificateFactory</code> will be used.
     *
     * @return <code>CertificateFactory</code> 
     */
    protected synchronized static CertificateFactory getCertificateFactory() 
	throws GeneralSecurityException {
	if (certFact == null) {
	    if (provider == null) {
		certFact = CertificateFactory.getInstance("X.509");
	    } else {
		certFact = CertificateFactory.getInstance("X.509", provider);
	    }
	}
	return certFact;
    }

    /**
     * Loads a X509 certificate from the specified input stream.
     * Input stream must contain DER-encoded certificate.
     *
     * @param in the input stream to read the certificate from.
     * @return <code>X509Certificate</code> the loaded certificate.
     * @exception GeneralSecurityException if certificate failed to load.
     */
    public static X509Certificate loadCertificate(InputStream in)
	throws GeneralSecurityException {
	return (X509Certificate)getCertificateFactory().generateCertificate(in);
    }

    /** 
     * Loads an X509 certificate from the specified file.  It
     * reads only what is in between the lines containing "BEGIN CERTIFICATE"
     * and "END".
     *
     * @param file the certificate file to load the certificate from.
     * @return <code>java.security.cert.X509Certificate</code> 
     *         the loaded certificate.
     * @exception IOException if I/O error occurs
     * @exception GeneralSecurityException if security problems occurs.
     */
    public static X509Certificate loadCertificate(String file) 
	throws IOException, GeneralSecurityException {

	if (file == null) {
	    throw new IllegalArgumentException("Certificate file is null");
	}

	boolean isCert = false;
	X509Certificate cert = null;

	BufferedReader reader = null;
	 
	String line       = null;
	StringBuffer buff = new StringBuffer();
	   
	reader = new BufferedReader(new FileReader(file));
	
	try {
	    while ((line = reader.readLine()) != null) {
		if (line.indexOf("BEGIN CERTIFICATE") != -1) {
		    isCert = true;
		} else if (isCert && line.indexOf("END CERTIFICATE") != -1) {
		    byte [] data = Base64.decode(buff.toString().getBytes());
		    cert = loadCertificate(new ByteArrayInputStream(data));
		} else if (isCert) {
		    buff.append(line);
		}
	    }
	} finally {
	    reader.close();
	}
	
	if (cert == null) {
	    throw new GeneralSecurityException("Certificate data not found.");
	}
	 
	return cert;
    }

    /**
     * Converts DN of the form "CN=foo, O=bar" into Globus 
     * format "/CN=foo/O=bar" (in the reverse way).
     * This function might return incorrect Globus-formatted ID when one of
     * the RDNs in the DN contains commas.
     * @see #toGlobusID(String, boolean)
     *
     * @param dn the DN to convert to Globus format.
     * @return the converted DN in Globus format.
     */
    public static String toGlobusID(String dn) {
	return toGlobusID(dn, true);
    }

    /**
     * Converts DN of the form "CN=foo, O=bar" into Globus 
     * format "/CN=foo/O=bar". This function might return incorrect 
     * Globus-formatted ID when one of the RDNs in the DN contains commas.
     *
     * @param dn the DN to convert to Globus format.
     * @param reverse the direction of the conversion.
     * @return the converted DN in Globus format.
     */
    public static String toGlobusID(String dn, boolean reverse) {
	if (dn == null) {
	    return null;
	}

	StringTokenizer tokens = new StringTokenizer(dn, ",");
	StringBuffer buf = new StringBuffer();
	String token;
	
	while(tokens.hasMoreTokens()) {
	    token = tokens.nextToken().trim();

	    if (reverse) {
		buf.append("/");
		buf.append(token);
	    } else {
		buf.insert(0, token);
		buf.insert(0, "/");
	    }
	}
	
	return buf.toString();
    }

    public static String toGlobusID(Principal name) {
	if (name instanceof X509Name) {
	    return X509NameHelper.toString((X509Name)name);
	} else {
	    return CertUtil.toGlobusID(name.getName());
	}
    }

    /**
     * Installs SecureRandom provider. 
     * This function is automatically called when this class is loaded.
     */
    public static void installSecureRandomProvider() {
	CoGProperties props = CoGProperties.getDefault();
        String providerName = props.getSecureRandomProvider();
        try {
	    logger.debug("Loading SecureRandom provider: " + providerName);
            Class providerClass = Class.forName(providerName);
            Security.insertProviderAt( (Provider)providerClass.newInstance(), 1 );
	} catch (Exception e) {
	    logger.debug("Unable to install PRNG. Using default PRNG.", e);
        }
    }

    /**
     * Generates a key pair of given algorithm and strength.
     *
     * @param algorithm the algorithm of the key pair.
     * @param bits the strength
     * @return <code>KeyPair</code> the generated key pair.
     * @exception GeneralSecurityException if something goes wrong.
     */
    public static KeyPair generateKeyPair(String algorithm, int bits) 
	throws GeneralSecurityException {
	KeyPairGenerator generator = null;
	if (provider == null) {
	    generator = KeyPairGenerator.getInstance(algorithm);
	} else {
	    generator = KeyPairGenerator.getInstance(algorithm, provider);
	}
	generator.initialize(bits);
	return generator.generateKeyPair();
    }


    // proxy utilies 

    /**
     * Determines if a specified certificate type indicates a GSI-2 or
     * GSI-3 proxy certificate.
     *
     * @param certType the certificate type to check.
     * @return true if certType is a GSI-2 or GSI-3 proxy, false 
     *         otherwise.
     */
    public static boolean isProxy(int certType) {
	return (isGsi2Proxy(certType) ||
		isGsi3Proxy(certType));
    }
    
    /**
     * Determines if a specified certificate type indicates a 
     * GSI-3 proxy certificate.
     *
     * @param certType the certificate type to check.
     * @return true if certType is a GSI-3 proxy, false 
     *         otherwise.
     */
    public static boolean isGsi3Proxy(int certType) {
	return (certType == GSIConstants.GSI_3_IMPERSONATION_PROXY ||
		certType == GSIConstants.GSI_3_INDEPENDENT_PROXY ||
		certType == GSIConstants.GSI_3_RESTRICTED_PROXY ||
		certType == GSIConstants.GSI_3_LIMITED_PROXY);
    }

    /**
     * Determines if a specified certificate type indicates a 
     * GSI-2 proxy certificate.
     *
     * @param certType the certificate type to check.
     * @return true if certType is a GSI-2 proxy, false 
     *         otherwise.
     */
    public static boolean isGsi2Proxy(int certType) {
	return (certType == GSIConstants.GSI_2_PROXY ||
		certType == GSIConstants.GSI_2_LIMITED_PROXY);
    }

    /**
     * Determines if a specified certificate type indicates a 
     * GSI-2 or GSI-3 limited proxy certificate.
     *
     * @param certType the certificate type to check.
     * @return true if certType is a GSI-2 or GSI-3 limited proxy, 
     *         false otherwise.
     */
    public static boolean isLimitedProxy(int certType) {
	return (certType == GSIConstants.GSI_3_LIMITED_PROXY ||
		certType == GSIConstants.GSI_2_LIMITED_PROXY);
    }

    /**
     * Determines if a specified certificate type indicates a 
     * GSI-2 or GSI-3 impersonation proxy certificate.
     *
     * @param certType the certificate type to check.
     * @return true if certType is a GSI-2 or GSI-3 impersonation proxy, 
     *         false otherwise.
     */
    public static boolean isImpersonationProxy(int certType) {
	return (certType == GSIConstants.GSI_3_IMPERSONATION_PROXY ||
		certType == GSIConstants.GSI_3_LIMITED_PROXY ||
		certType == GSIConstants.GSI_2_LIMITED_PROXY ||
		certType == GSIConstants.GSI_2_PROXY);
    }
    
    /**
     * Returns a string description of a specified proxy
     * type.
     *
     * @param proxyType the proxy type to get the string 
     *        description of.
     * @return the string description of the proxy type.
     */
    public static String getProxyTypeAsString(int proxyType) {
	switch(proxyType) {
	case GSIConstants.GSI_3_IMPERSONATION_PROXY:
	    return "Proxy draft compliant impersonation proxy";
	case GSIConstants.GSI_3_INDEPENDENT_PROXY:
	    return "Proxy draft compliant independent proxy";
	case GSIConstants.GSI_3_LIMITED_PROXY:
	    return "Proxy draft compliant limited proxy";
	case GSIConstants.GSI_3_RESTRICTED_PROXY:
	    return "Proxy draft compliant restricted proxy";
	case GSIConstants.GSI_2_PROXY:
	    return "full legacy globus proxy";
	case GSIConstants.GSI_2_LIMITED_PROXY:
	    return "limited legacy globus proxy";
	default:
	    return "not a proxy";
	}
    }

    /**
     * Checks if GSI-3 mode is enabled.
     *
     * @return true if <I>"org.globus.gsi.version"</I> system property
     *         is set to "3". Otherwise, false.
     */
    public static boolean isGsi3Enabled() {
	String ver = System.getProperty("org.globus.gsi.version");
	return (ver != null && ver.equals("3"));
    }

    // CRL Utilities FIXME - move to separate class ?
    public static X509CRL loadCrl(String file) 
	throws IOException, GeneralSecurityException {

	if (file == null) {
	    throw new IllegalArgumentException("CRL file is null");
	}

	boolean isCrl = false;
	X509CRL crl = null;

	BufferedReader reader = null;
	 
	String line       = null;
	StringBuffer buff = new StringBuffer();
	   
	reader = new BufferedReader(new FileReader(file));
	
	try {
	    while ((line = reader.readLine()) != null) {
		if (line.indexOf("BEGIN X509 CRL") != -1) {
		    isCrl = true;
		} else if (isCrl && line.indexOf("END X509 CRL") != -1) {
		    byte [] data = Base64.decode(buff.toString().getBytes());
		    crl = loadCrl(new ByteArrayInputStream(data));
		} else if (isCrl) {
		    buff.append(line);
		}
	    }
	} finally {
	    reader.close();
	}
	
	if (crl == null) {
	    throw new GeneralSecurityException("CRL data not found.");
	}
	
	return crl;
    }

    public static X509CRL loadCrl(InputStream in)
	throws GeneralSecurityException {
	return (X509CRL)getCertificateFactory().generateCRL(in);
    }
}
