/*
This file is licensed under the terms of the Globus Toolkit Public
License, found at http://www.globus.org/toolkit/download/license.html.
*/
package org.globus.gsi.gssapi.auth;

import org.gridforum.jgss.ExtendedGSSManager;

import org.ietf.jgss.GSSManager;
import org.ietf.jgss.GSSContext;
import org.ietf.jgss.GSSName;
import org.ietf.jgss.GSSException;
import org.ietf.jgss.GSSCredential;

import org.apache.log4j.Logger;

/**
 * Implements a simple identity authorization mechanism.
 * The given identify is compared with the peer's identity.
 */
public class IdentityAuthorization 
    extends GSSAuthorization {

    private static Logger logger =
        Logger.getLogger(IdentityAuthorization.class.getName());

    protected String _identity;
    
    /**
     * Constructor used by superclasses.
     */
    protected IdentityAuthorization() {
    }

    /**
     * Creates a new instance of this class with given
     * expected identity.
     *
     * @param identity the expected identity. Must not be null.
     */
    public IdentityAuthorization(String identity) {
	setIdentity(identity);
    }

    /**
     * Sets the expected identity for the authorization 
     * check.
     *
     * @param identity the expected identity. Must not be null.
     */
    public void setIdentity(String identity) {
	if (identity == null) {
	    throw new IllegalArgumentException("Identity cannot be null");
	}
	_identity = identity;
    }
    
    /**
     * Returns the expected identity.
     *
     * @return the expected identity.
     */
    public String getIdentity() {
	return _identity;
    }

    public GSSName getExpectedName(GSSCredential cred, String host) 
	throws GSSException {
        GSSManager manager = ExtendedGSSManager.getInstance();
        return manager.createName(_identity, null);
    }

    /**
     * Performs identity authorization. The given identity is compared
     * with the peer's identity.
     *
     * @param context the security context
     * @param socket the socket connected to the peer.
     * @exception AuthorizationException if the peer's
     *            identity does not match the expected identity.
     */
    public void authorize(GSSContext context, String host)
	throws AuthorizationException {
	logger.info("Authorization: IDENTITY");

	try {
	    GSSName expected = getExpectedName(null, host);
	
	    GSSName target = null;
	    if (context.isInitiator()) {
		target = context.getTargName();
	    } else {
		target = context.getSrcName();
	    }
	    
	    if (!expected.equals(target)) {
		generateAuthorizationException(expected, target);
	    }
	} catch (GSSException e) {
	    throw new AuthorizationException("Authorization failure", e);
	}
    }
    
}
