/*
This file is licensed under the terms of the Globus Toolkit Public
License, found at http://www.globus.org/toolkit/download/license.html.
*/
package org.globus.net;

import java.io.IOException;

import org.globus.common.CoGProperties;

/**
 * This class manages the port ranges. It keeps track of which
 * ports are used and which ones are open.
 */
public class PortRange {

    protected static final byte UNUSED = 0;
    protected static final byte USED   = 1;
    
    private boolean portRange = false;
    private int minPort, maxPort;
    private byte [] ports;
  
    private static PortRange defaultPortRange = null;
  
    protected PortRange() {
	init();
    }
  
    /**
     * Returns the default instance of this class.
     * If the tcp.port.range property is set, the class
     * will be initialized with the specified port ranges.
     *
     * @return PortRange the default instace of this class.
     */
    public static synchronized PortRange getInstance() {
	if (defaultPortRange == null) {
	    defaultPortRange = new PortRange();
	}
	return defaultPortRange;
    }
    
    /**
     * Checks if the port range is set.
     *
     * @return true if the port range is set, false otherwise.
     */
    public boolean isEnabled() {
	return portRange;
    }
  
    /**
     * Returns first available port.
     *
     * @param lastPortNumber port number to start finding the next 
     *                       available port from. Set it to 0 if
     *                       called initialy.
     * @return the next available port number from the lastPortNumber.
     * @exception IOException if there is no more free ports available or
     *            if the lastPortNumber is incorrect.
     */
    public int getFreePort(int lastPortNumber) 
	throws IOException {
	int id = 0;
	if (lastPortNumber != 0) {
	    id = lastPortNumber - minPort;
	    if (id < 0) {
		throw new IOException("Port number out of range.");
	    }
	}
	for(int i=id;i<ports.length;i++) {
	    if (ports[i] == USED) continue;
	    return minPort+i;
	}
	throw new IOException("No free ports available.");
    }
  
    /**
     * Sets the port number as used.
     *
     * @param portNumber port number 
     */
    public void setUsed(int portNumber) {
	setPort(portNumber, USED);
    }
  
    /**
     * Releases or frees the port number.
     * (Mark it as unused)
     * 
     * @param portNumber port number
     */
    public void free(int portNumber) {
	setPort(portNumber, UNUSED);
    }
    
    private void setPort(int portNumber, byte type) {
	int id = portNumber - minPort;
	if (id < 0) {
	    throw new IllegalArgumentException("Port number out of range: " + portNumber);
	}
	ports[id] = type;
    }
    
    private void init() {
	init( CoGProperties.getDefault().getTcpPortRange() );
    }
  
    private void init(String portRangeStr) {
	portRange = false;
	
	if (portRangeStr == null) return ;
	
	int pos = portRangeStr.indexOf(",");
	if (pos == -1) {
	    throw new IllegalArgumentException("Missing comma in the port range property: " +
					       portRangeStr);
	}
	
	int min, max;
	
	try {
	    min = Integer.parseInt(portRangeStr.substring(0, pos).trim());
	} catch(Exception e) {
	    throw new IllegalArgumentException("The minimum port range value is invalid: " + 
					       e.getMessage());
	}
	
	try {
	    max = Integer.parseInt(portRangeStr.substring(pos+1).trim());
	} catch(Exception e) {
	    throw new IllegalArgumentException("The maximum port range value is invalid: " + 
					       e.getMessage());
	}
    
	if (min >= max) {
	    throw new IllegalArgumentException("The minimum port range value is greater then " + 
					       "the maximum port range value.");
	}
    
	minPort = min;
	maxPort = max;
	ports     = new byte[ maxPort-minPort ];
	portRange = true;    
    }
    
}
