/*
This file is licensed under the terms of the Globus Toolkit Public
License, found at http://www.globus.org/toolkit/download/license.html.
*/
package org.globus.security.gridmap;

import java.util.Map;
import java.util.Vector;
import java.util.HashMap;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.io.IOException;
import java.io.File;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.io.InputStream;

import org.globus.util.QuotedStringTokenizer;


//TODO: globusID lookups in map are case sensitive.

public class GridMap {

    private static final String COMMENT_CHARS = "#";

    protected Map map;

    // the file the grim map was loaded from
    private File file;
    // last time the file was modified
    private long lastModified;

    /**
     * Loads grid map definition from a given 
     * file. 
     * 
     * @param file grid map file
     * @exception IOException in case of I/O or 
     *            parsing error.
     */
    public void load(String file)
	throws IOException {
	load(new File(file));
    }

    /**
     * Loads grid map definition from a given 
     * file. 
     * 
     * @param file grid map file
     * @exception IOException in case of I/O or 
     *            parsing error.
     */
    public void load(File file)
	throws IOException {
	InputStream in = null;
	try {
	    in = new FileInputStream(file);
	    this.file = file;
	    this.lastModified = file.lastModified();
	    load(in);
	} finally {
	    if (in != null) {
		try { in.close(); } catch(Exception e) {}
	    }
	}
    }

    /**
     * Reloads the gridmap from a file only if the gridmap
     * was initially loaded using the {@link #load(File) 
     * load} or {@link #load(String) load} functions.
     * The file will only be reloaded if it has changed
     * since the last time. 
     *
     * @exception IOException in case of I/O error during
     *            reload.
     */
    public void refresh() 
	throws IOException {
	if (this.file == null) {
	    return;
	}
	if (this.file.lastModified() != this.lastModified) {
	    load(this.file);
	}
    }
    
    /**
     * Loads grid map file definition from a given
     * input stream. The input stream is not closed
     * in case of an error.
     *
     * @param input input stream containing grid map 
     *        definitions.
     * @exception IOException in case of I/O error or
     *            parsing error.
     */
    public void load(InputStream input) 
	throws IOException {
	String line;
	
	BufferedReader reader = 
	    new BufferedReader(new InputStreamReader(input));

	Map localMap = new HashMap();
	GridMapEntry entry;
	QuotedStringTokenizer tokenizer;
	StringTokenizer idTokenizer;
	
	while( (line = reader.readLine()) != null) {
	    line = line.trim();
	    if ( (line.length() == 0) ||
		 ( COMMENT_CHARS.indexOf(line.charAt(0)) != -1) ) {
		continue;
	    }
	    
	    tokenizer = new QuotedStringTokenizer(line);

	    String globusID = null;

	    if (tokenizer.hasMoreTokens()) {
		globusID = tokenizer.nextToken();
	    } else {
		throw new IOException("Globus ID not defined: " + line);
	    }

	    String userIDs = null;
	    
	    if (tokenizer.hasMoreTokens()) {
		userIDs = tokenizer.nextToken();
	    } else {
		throw new IOException("User ID mapping missing: " + line);
	    }

	    idTokenizer = new StringTokenizer(userIDs, ",");
	    String [] ids = new String [ idTokenizer.countTokens() ];
	    int i = 0;
	    while(idTokenizer.hasMoreTokens()) {
		ids[i++] = idTokenizer.nextToken();
	    }
	    
	    entry = (GridMapEntry)localMap.get(globusID);
	    if (entry == null) {
		entry = new GridMapEntry();
		entry.setGlobusID(globusID);
		entry.setUserIDs(ids);
		localMap.put(globusID, entry);
	    } else {
		entry.addUserIDs(ids);
	    }
	}
	
	map = localMap;
    }

    /**
     * Returns first local user name mapped to the specified 
     * globusID.
     * 
     * @param globusID globusID
     * @return local user name for the specified globusID.
     *         Null if the globusID is not mapped
     *         to a local user name. 
     */
    public String getUserID(String globusID) {
	String [] ids = getUserIDs(globusID);
	if (ids != null && ids.length > 0) {
	    return ids[0];
	} else {
	    return null;
	}
    }
    
    /**
     * Returns local user names mapped to the specified
     * globusID.
     *
     * @param globusID globusID
     * @return array of local user names for the specified globusID.
     *         Null if the globusID is not mapped
     *         to any local user name.
     */
    public String[] getUserIDs(String globusID) {
	if (globusID == null) {
	    throw new IllegalArgumentException("globusID is null");
	}
	
	if (map == null) {
	    // not initialized
	    // if not running as root, then return own username
	    // this allows someone to run as themselves without
	    // having to have a gridmap file.
	    String user = System.getProperty("user.name");
	    if (user == null) return null;
	    String tmpUser = user.toLowerCase();
	    if (tmpUser.equals("root") ||
		tmpUser.equals("administrator")) {
		return null;
	    } else {
		return new String[] {user};
	    }
	} 
	    
	GridMapEntry entry = (GridMapEntry)map.get(globusID);
	return (entry == null) ? null : entry.getUserIDs();
    }

    /**
     * Checks if a given globus ID is associated with given
     * local user account.
     *
     * @param globusID globus ID
     * @param userID userID
     * @return true if globus ID is associated with given local
     *         user account, false, otherwise.
     */
    public boolean checkUser(String globusID, String userID) {
	if (globusID == null) {
	    throw new IllegalArgumentException("globusID is null");
	}
	if (userID == null) {
	    throw new IllegalArgumentException("userID is null");
	}
	
	if (map == null) {
	    // not initialized
	    // if not running as root, then return own username
	    // this allows someone to run as themselves without
	    // having to have a gridmap file.
	    String user = System.getProperty("user.name");
	    if (user == null) return false;
	    String tmpUser = user.toLowerCase();
	    if (tmpUser.equals("root") ||
		tmpUser.equals("administrator")) {
		return false;
	    } else {
		return user.equalsIgnoreCase(userID);
	    }
	}

	GridMapEntry entry = (GridMapEntry)map.get(globusID);
	return (entry == null) ? false : entry.containsUserID(userID);
    }

    /**
     * Returns globus ID associated with the
     * specified local user name.
     *
     * @param userID local user name
     * @return associated globus ID, null
     *         if there is not any.
     */
    public String getGlobusID(String userID) {
	if (userID == null) {
	    throw new IllegalArgumentException("userID is null");
	}
	
	if (map == null) {
	    return null;
	}

	Iterator iter = map.entrySet().iterator();
	Map.Entry mapEntry;
	GridMapEntry entry;
	while(iter.hasNext()) {
	    mapEntry = (Map.Entry)iter.next();
	    entry = (GridMapEntry)mapEntry.getValue();
	    if (entry.containsUserID(userID)) {
		return entry.getGlobusID();
	    }
	}
	return null;
    }

    /**
     * Returns all globus IDs associated with the
     * specified local user name.
     *
     * @param userID local user name
     * @return associated globus ID, null
     *         if there is not any.
     */
    public String[] getAllGlobusID(String userID) {
        if (userID == null) {
	    throw new IllegalArgumentException("userID is null");
	}
	
        if (map == null) {
            return null;
        }

        Vector v = new Vector();

        Iterator iter = map.entrySet().iterator();
        Map.Entry mapEntry;
        GridMapEntry entry;
        while(iter.hasNext()) {
            mapEntry = (Map.Entry)iter.next();
            entry = (GridMapEntry)mapEntry.getValue();
            if (entry.containsUserID(userID)) {
                v.add(entry.getGlobusID());
           }
        }

        // create array of strings and add values back in
        if(v.size() == 0) {
            return null;
        }

        String idS[] = new String[v.size()];
        for(int ctr = 0; ctr < v.size(); ctr++) {
            idS[ctr] = (String) v.elementAt(ctr);
        }

        return idS;
    }

    public void map(String globusID, String userID) {
	if (globusID == null) {
	    throw new IllegalArgumentException("globusID is null");
	}
	if (userID == null) {
	    throw new IllegalArgumentException("userID is null");
	}
	
	if (map == null) {
	    map = new HashMap();
	}
	
	GridMapEntry entry = (GridMapEntry)map.get(globusID);
	if (entry == null) {
	    entry = new GridMapEntry();
	    entry.setGlobusID(globusID);
	    entry.setUserIDs(new String [] {userID});
	    map.put(globusID, entry);
	} else {
	    entry.addUserID(userID);
	}
    }

    static class GridMapEntry {
	String globusID;
	String[] userIDs;

	public String getFirstUserID() {
	    return userIDs[0];
	}

	public String[] getUserIDs() {
            return userIDs;
        }

	public String getGlobusID() {
	    return globusID;
	}

	public void setGlobusID(String globusID) {
	    this.globusID = globusID;
	}
	
	public void setUserIDs(String [] userIDs) {
	    this.userIDs = userIDs;
	}

	public boolean containsUserID(String userID) {
	    if (userID == null) {
		return false;
	    }
	    for (int i=0;i<userIDs.length;i++) {
		if (userIDs[i].equalsIgnoreCase(userID)) {
		    return true;
		}
	    }
	    return false;
	}

	public void addUserID(String userID) {
	    if (containsUserID(userID)) return;
	    String [] ids = new String[ userIDs.length + 1 ];
	    System.arraycopy(userIDs, 0, ids, 0, userIDs.length);
	    ids[userIDs.length] = userID;
	    userIDs = ids;
	}
	
	public void addUserIDs(String [] userIDs) {
	    for (int i=0;i<userIDs.length;i++) {
		addUserID(userIDs[i]);
	    }
	}

    }

}
