/*
This file is licensed under the terms of the Globus Toolkit Public
License, found at http://www.globus.org/toolkit/download/license.html.
*/
package org.globus.tomcat.catalina.net;

import java.io.IOException;
import java.net.Socket;
import java.net.InetAddress;
import java.net.ServerSocket;

import org.globus.common.ChainedIOException;
import org.globus.security.gridmap.GridMap;
import org.globus.gsi.GlobusCredential;
import org.globus.gsi.GlobusCredentialException;
import org.globus.gsi.TrustedCertificates;
import org.globus.gsi.GSIConstants;
import org.globus.gsi.gssapi.GlobusGSSCredentialImpl; // impl. specific
import org.globus.gsi.gssapi.GSSConstants;
import org.globus.gsi.gssapi.net.GssSocketFactory;

import org.gridforum.jgss.ExtendedGSSManager;
import org.gridforum.jgss.ExtendedGSSContext;

import org.ietf.jgss.GSSCredential;
import org.ietf.jgss.GSSManager;
import org.ietf.jgss.GSSException;

import org.apache.log4j.Logger;

// only dependency on tomcat lib
import org.apache.catalina.net.ServerSocketFactory;

public class GSIServerSocketFactory 
    implements ServerSocketFactory {

    private String serverCert   = "/etc/grid-security/hostcert.pem";
    private String serverKey    = "/etc/grid-security/hostkey.pem";
    private String serverProxy;
    private String caCertDir    = "/etc/grid-security/certificates";
    private String gridMapFile  = "/etc/grid-security/grid-mapfile";
    private boolean sslMode;
    
    private static Logger logger = 
	Logger.getLogger(GSIServerSocketFactory.class);

    public GSIServerSocketFactory() {
	sslMode = false;
    }
    
    public void setProxy(String serverProxy) {
	this.serverProxy = serverProxy;
    }

    public String getProxy() {
        return serverProxy;
    }

    public void setCert(String serverCert) {
	this.serverCert = serverCert;
    }

    public String getCert() {
        return serverCert;
    }

    public void setKey(String serverKey) {
	this.serverKey = serverKey;
    }

    public String getKey() {
        return serverKey;
    }

    public void setCacertdir(String caCertDir) {
	this.caCertDir = caCertDir;
    }

    public String getCacertdir() {
        return caCertDir;
    }

    public void setGridMap(String file) {
	this.gridMapFile = file;
    }

    public String getGridMap() {
	return gridMapFile;
    }

    public void setSslMode(boolean sslMode) {
	this.sslMode = sslMode;
    }

    public boolean getSslMode() {
	return sslMode;
    }

    // ------------------------------------------

    /**
     * Creates a secure server socket on a specified port with default
     * user credentials. A port of <code>0</code> creates a socket on 
     * any free port or if the tcp.port.range system property is set 
     * it creates a socket within the specified port range.
     * <p>
     * The maximum queue length for incoming connection indications (a
     * request to connect) is set to <code>50</code>. If a connection
     * indication arrives when the queue is full, the connection is refused.
     * <p>
     *
     * @param      port  the port number, or <code>0</code> to use any
     *                   free port or if the tcp.port.range property set
     *                   to use any available port within the specified port
     *                   range.
     * @exception  IOException  
     *              if an I/O error occurs when opening the socket.
     */
    public ServerSocket createSocket(int port) 
	throws IOException {
	return createSocket(port, 50);
    }
    
    /**
     * Creates a secure server socket on a specified port with default
     * user credentials. A port of <code>0</code> creates a socket on
     * any free port or if the tcp.port.range system property is set
     * it creates a socket within the specified port range.
     * <p>
     * The maximum queue length for incoming connection indications (a
     * request to connect) is set to the <code>backlog</code> parameter. If
     * a connection indication arrives when the queue is full, the
     * connection is refused.
     *
     * @param      port  the port number, or <code>0</code> to use any
     *                   free port or if the tcp.port.range property set
     *                   to use any available port within the specified port
     *                   range.
     * @param      backlog  
     *              the maximum length of the queue.
     * @exception  IOException  
     *              if an I/O error occurs when opening the socket.
     */
    public ServerSocket createSocket(int port, int backlog) 
	throws IOException {
	return createSocket(port, backlog, null);
    }
    
    /**
     * Creates a secure server socket on a specified port with default
     * user credentials. A port of <code>0</code> creates a socket on
     * any free port or if the tcp.port.range system property is set
     * it creates a socket within the specified port range.
     * <p>
     * The maximum queue length for incoming connection indications (a
     * request to connect) is set to the <code>backlog</code> parameter. If
     * a connection indication arrives when the queue is full, the
     * connection is refused.
     *
     * @param      port  the port number, or <code>0</code> to use any
     *                   free port or if the tcp.port.range property set
     *                   to use any available port within the specified port
     *                   range.
     * @param      backlog  
     *              the maximum length of the queue.
     * @param      bindAddr the local InetAddress the server will bind to.
     * @exception  IOException  
     *              if an I/O error occurs when opening the socket.
     */
    public ServerSocket createSocket(int port, 
				     int backlog,
				     InetAddress bindAddr)
	throws IOException {

	GSSCredential gssCred = null;
	GlobusCredential cred = null;

	logger.info("Loading credentials");
	try {
	    if (serverProxy != null && !serverProxy.equals("")) {
		if (logger.isInfoEnabled()) {
		    logger.info("Server Proxy: " + serverProxy);
		}
		cred = new GlobusCredential(serverProxy);
	    } else {
		if (logger.isInfoEnabled()) {
		    logger.info("Server Certificate: " + serverCert);
		    logger.info("Server Key: " + serverKey);
		}
		cred = new GlobusCredential(serverCert, serverKey);
	    } 

	    gssCred = new GlobusGSSCredentialImpl(cred, 
						  GSSCredential.ACCEPT_ONLY);
	} catch (GlobusCredentialException e) {
	    throw new ChainedIOException("Failed to load credentials", e);
	} catch (GSSException e) {
	    throw new ChainedIOException("Failed to load credentials", e);
	}
	
	GridMap gridMap = new GridMap();

	if (logger.isInfoEnabled()) {
	    logger.info("Loading gridmap file:" + gridMapFile);
	}
	
	gridMap.load(gridMapFile);
	
	if (logger.isInfoEnabled()) {
	    logger.info("CA certificate directory: " + caCertDir);
	}
	
	TrustedCertificates trustedCerts = 
	    TrustedCertificates.load(caCertDir);
	
	GSIServerSocket serverSocket = 
	    new GSIServerSocket(port, backlog, bindAddr);
	
	serverSocket.setCredentials(gssCred);
	serverSocket.setGridMap(gridMap);
	serverSocket.setTrustedCertificates(trustedCerts);

	return serverSocket;
    }
    
    class GSIServerSocket extends ServerSocket {

	private GridMap _gridMap;
	private TrustedCertificates _trustedCerts;

	private GSSCredential _credentials;
	private GssSocketFactory _factory;
	private GSSManager _manager;

	public GSIServerSocket(int port, int backlog, 
			       InetAddress bindAddr)
	    throws IOException {
	    super(port, backlog, bindAddr);
	    
	    this._factory = GssSocketFactory.getDefault();
	    this._manager = ExtendedGSSManager.getInstance();
	}

	public void setTrustedCertificates(TrustedCertificates trustedCerts) {
	    this._trustedCerts = trustedCerts;
	}

	public void setGridMap(GridMap gridMap) {
	    this._gridMap = gridMap;
	}
	
	public void setCredentials(GSSCredential creds) {
	    this._credentials = creds;
	}

	public Socket accept()
	    throws IOException {

	    Socket s = super.accept();
	 
	    ExtendedGSSContext context = null;

	    try {
		context = 
		    (ExtendedGSSContext)this._manager.createContext(this._credentials);
	    
		if (getSslMode()) {
		    logger.info("Using SSL mode");
		    context.setOption(GSSConstants.GSS_MODE,
				      GSIConstants.MODE_SSL);
		}

		if (this._trustedCerts != null) {
		    logger.info("Setting trusted certificates");
		    context.setOption(GSSConstants.TRUSTED_CERTIFICATES,
				      this._trustedCerts);
		}
	    } catch (GSSException e) {
		throw new ChainedIOException("Failed to init GSS context", e);
	    }

	    GSISocket gsiSocket = new GSISocket(s, context);
	    
	    // server socket
	    gsiSocket.setUseClientMode(false);

	    // refresh grid map
	    this._gridMap.refresh();

	    // set grid map
	    gsiSocket.setGridMap(this._gridMap);

	    return gsiSocket;
	}
    }
    
}
